/* Aranha Parent process manager
 *
 * arch-tag: e70959ed-e4a2-4a14-9ac3-dcc3aa35006b
 *
 * Copyright Daniel Silverstone <dsilvers@digital-scurf.org>
 * Copyright (C) 2004-2005 Pepperfish Ltd
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "config.h"
#include "configurator.h"
#include "lua.h"
#include "lauxlib.h"
#include "parentchild.h"
#include "interpreters.h"
#include "utils.h"

#include <unistd.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <signal.h>

#include <string.h>

lua_State** states;

static int 
L_sleep( lua_State *L )
{
  if( lua_gettop(L) != 1 || 
      lua_isnumber(L, 1) == 0 ) {
    lua_pushstring( L, "sleep expects a single numeric argument" );
    lua_error(L);
  }
  sleep(lua_tonumber(L,1));
  return 0;
}

static lua_State *
prepare_state(char *preload)
{
  lua_State *L = prepare_full_interpreter(ICTX_ARANHA);
  lua_settop(L,0);
  
  lua_pushcclosure( L, L_sleep, 0 );
  lua_setglobal( L, "sleep" );
  
  if( preload ) {
    lua_pushliteral(L, "_TRACEBACK");
    lua_gettable(L, LUA_GLOBALSINDEX);
    if( luaL_loadfile(L, preload) ) {
      BAIL(lua_tostring(L,-1));
    }
    
    if( lua_pcall(L, 0, 0, -2) > 0 ) {
      BAIL(lua_tostring(L,-1));
    }
  }
  lua_settop(L,0);
  
  {
#include "startup/aranha-post.cinc"
  }
  
  lua_settop(L,0);
  
  return L;
}

static int perchild;
static int reqperstate;

static void
InitialiseStates( void ) 
{
  int i;
  char *preload;
  
  if( get_config_int("statecount", &perchild) )
    perchild = 1;
  if( get_config_string("preload", &preload) )
    preload = NULL;
  if( get_config_int("reqperstate", &reqperstate) )
    reqperstate = 1;
  
  states = malloc( (sizeof( lua_State* )) * perchild );
  
  for( i = 0; i < perchild; ++i ) {
    states[i] = prepare_state(preload);
  }  
}

static int kid_died = 0;
void sig_child(int sig) { kid_died = 1; }

void 
PrepareParent() 
{
  InitialiseStates();
  /* set up signal handlers and generally get ready for having kids... */
  signal( SIGCHLD, sig_child );
}

static int kid_count = 0;

static void 
StartChild()
{
  /* Purpose: fork() a child, return in the parent, 
   * start ChildLoop in the kid */
  int res = fork();
  if( res == -1 ) return;
  if( res == 0 ) {
    /* Child */
    PrepareChild();
    ChildLoop(perchild, reqperstate);
  } else {
    /* Parent */
    kid_count++;
  }
}

void 
ParentLoop()
{
  int temp;
  int maxchildren;
  
  if( get_config_int( "children", &maxchildren ) )
    maxchildren = 1;
  while(1) {
    if( kid_died ) {
      /* Very simple reaper */
      kid_died = 0;
      while( wait3( &temp, WNOHANG, NULL ) > 0 ) { kid_count--; }
    }
    if( kid_count < maxchildren )
      StartChild();
    if( kid_died ) {
      /* Very simple reaper */
      kid_died = 0;
      while( wait3( &temp, WNOHANG, NULL ) > 0 ) { kid_count--; }
    }
    /* Got the right number of kids */
    if( kid_count < maxchildren )
      sleep( 1 );
    else
      /* Even if we miss the child, we'll catch it after a minute */
      sleep( 60 );
  }
}
