/*
** $Id: lparser.c,v 2.27 2005/05/17 19:49:15 roberto Exp $
** Lua Parser
** See Copyright Notice in lua.h
*/


#include <string.h>
#include <stdio.h>

#define lparser_c
#define LUA_CORE

#include "lua.h"

#include "lcode.h"
#include "ldebug.h"
#include "ldo.h"
#include "lfunc.h"
#include "llex.h"
#include "lmem.h"
#include "lobject.h"
#include "lopcodes.h"
#include "lparser.h"
#include "lstate.h"
#include "lstring.h"




#define hasmultret(k)		((k) == VCALL || (k) == VVARARG)

#define getlocvar(fs, i)	((fs)->f->locvars[(fs)->actvar[i]])

#define luaY_checklimit(fs,v,l,m)	if ((v)>(l)) errorlimit(fs,l,m)


/*
** nodes for block list (list of active blocks)
*/
typedef struct BlockCnt {
  struct BlockCnt *previous;  /* chain */
  int breaklist;  /* list of jumps out of this loop */
  lu_byte nactvar;  /* # active locals outside the breakable structure */
  lu_byte upval;  /* true if some variable in the block is an upvalue */
  lu_byte isbreakable;  /* true if `block' is a loop */
} BlockCnt;

/*
** prototype for the body() function
*/
static void body (LexState *ls, expdesc *e, int needself, int line);

static void localstat (LexState *ls);


/*
** prototypes for recursive non-terminal functions
*/
static void chunk (LexState *ls);
static void expr (LexState *ls, expdesc *v);

struct LockstepControl {
  int loop;  /* The index of the (T)FORLOOP op */
  int bodyj;  /* The index of the JMP to the body (loop or loop + 1) */
  int prepj;  /* The index of the FORPREP (or JMP) op */
  int maxnvars;  /* Largest number of nvars encountered in a TFORLOOP */
  int tempbase;  /* temporary base to use for TFORLOOPS */
};

static void forclause (LexState *ls, int line, int breakable, struct LockstepControl *ctl,
                     void(*bodyfunc)(LexState *ls, expdesc *t), expdesc *t);


static void next (LexState *ls) {
  ls->lastline = ls->linenumber;
  if (ls->lookahead.token != TK_EOS) {  /* is there a look-ahead token? */
    ls->t = ls->lookahead;  /* use this one */
    ls->lookahead.token = TK_EOS;  /* and discharge it */
  }
  else
    ls->t.token = luaX_lex(ls, &ls->t.seminfo);  /* read next token */
}


static void lookahead (LexState *ls) {
  lua_assert(ls->lookahead.token == TK_EOS);
  ls->lookahead.token = luaX_lex(ls, &ls->lookahead.seminfo);
}


static void anchor_token (LexState *ls) {
  if (ls->t.token == TK_NAME || ls->t.token == TK_STRING) {
    TString *ts = ls->t.seminfo.ts;
    luaX_newstring(ls, getstr(ts), ts->tsv.len);
  }
}


static void error_expected (LexState *ls, int token) {
  luaX_syntaxerror(ls,
      luaO_pushfstring(ls->L, LUA_QS " expected", luaX_token2str(ls, token)));
}


static void errorlimit (FuncState *fs, int limit, const char *what) {
  const char *msg = (fs->f->linedefined == 0) ?
    luaO_pushfstring(fs->L, "main function has more than %d %s", limit, what) :
    luaO_pushfstring(fs->L, "function at line %d has more than %d %s",
                            fs->f->linedefined, limit, what);
  luaX_lexerror(fs->ls, msg, 0);
}


static int testnext (LexState *ls, int c) {
  if (ls->t.token == c) {
    next(ls);
    return 1;
  }
  else return 0;
}


static void check (LexState *ls, int c) {
  if (ls->t.token != c)
    error_expected(ls, c);
}

static void checknext (LexState *ls, int c) {
  check(ls, c);
  next(ls);
}


#define check_condition(ls,c,msg)	{ if (!(c)) luaX_syntaxerror(ls, msg); }



static void check_match (LexState *ls, int what, int who, int where) {
  if (!testnext(ls, what)) {
    if (where == ls->linenumber)
      error_expected(ls, what);
    else {
      luaX_syntaxerror(ls, luaO_pushfstring(ls->L,
             LUA_QS " expected (to close " LUA_QS " at line %d)",
              luaX_token2str(ls, what), luaX_token2str(ls, who), where));
    }
  }
}


static TString *str_checkname (LexState *ls) {
  TString *ts;
  check(ls, TK_NAME);
  ts = ls->t.seminfo.ts;
  next(ls);
  return ts;
}


static void init_exp (expdesc *e, expkind k, int i) {
  e->f = e->t = NO_JUMP;
  e->k = k;
  e->info = i;
}


static void codestring (LexState *ls, expdesc *e, TString *s) {
  init_exp(e, VK, luaK_stringK(ls->fs, s));
}


static void checkname(LexState *ls, expdesc *e) {
  codestring(ls, e, str_checkname(ls));
}


static int registerlocalvar (LexState *ls, TString *varname) {
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  int oldsize = f->sizelocvars;
  luaM_growvector(ls->L, f->locvars, fs->nlocvars, f->sizelocvars,
                  LocVar, SHRT_MAX, "too many local variables");
  while (oldsize < f->sizelocvars) f->locvars[oldsize++].varname = NULL;
  f->locvars[fs->nlocvars].varname = varname;
  luaC_objbarrier(ls->L, f, varname);
  return fs->nlocvars++;
}


#define new_localvarliteral(ls,v,n) \
  new_localvar(ls, luaX_newstring(ls, "" v, (sizeof(v)/sizeof(char))-1), n)


static void new_localvar (LexState *ls, TString *name, int n) {
  FuncState *fs = ls->fs;
  luaY_checklimit(fs, fs->nactvar+n+1, LUAI_MAXVARS, "local variables");
  if (fs->nactvar+n+1 > fs->nregvar) fs->nregvar = fs->nactvar+n+1;
  fs->actvar[fs->nactvar+n] = cast(unsigned short, registerlocalvar(ls, name));
}


static void adjustlocalvars (LexState *ls, int nvars) {
  FuncState *fs = ls->fs;
  fs->nactvar += nvars;
  for (; nvars; nvars--) {
    getlocvar(fs, fs->nactvar - nvars).startpc = fs->pc;
  }
}


static void removevars (LexState *ls, int tolevel) {
  FuncState *fs = ls->fs;
  while (fs->nactvar > tolevel)
    getlocvar(fs, --fs->nactvar).endpc = fs->pc;
  fs->nregvar = tolevel;
}


static int indexupvalue (FuncState *fs, TString *name, expdesc *v) {
  int i;
  Proto *f = fs->f;
  int oldsize = f->sizeupvalues;
  for (i=0; i<f->nups; i++) {
    if (fs->upvalues[i].k == v->k && fs->upvalues[i].info == v->info) {
      lua_assert(f->upvalues[i] == name);
      return i;
    }
  }
  /* new one */
  luaY_checklimit(fs, f->nups + 1, LUAI_MAXUPVALUES, "upvalues");
  luaM_growvector(fs->L, f->upvalues, f->nups, f->sizeupvalues,
                  TString *, MAX_INT, "");
  while (oldsize < f->sizeupvalues) f->upvalues[oldsize++] = NULL;
  f->upvalues[f->nups] = name;
  luaC_objbarrier(fs->L, f, name);
  lua_assert(v->k == VLOCAL || v->k == VUPVAL);
  fs->upvalues[f->nups].k = cast(lu_byte, v->k);
  fs->upvalues[f->nups].info = cast(lu_byte, v->info);
  return f->nups++;
}


static int searchvar (FuncState *fs, TString *n) {
  int i;
  for (i=fs->nactvar-1; i >= 0; i--) {
    if (n == getlocvar(fs, i).varname)
      return i;
  }
  return -1;  /* not found */
}


static void markupval (FuncState *fs, int level) {
  BlockCnt *bl = fs->bl;
  while (bl && bl->nactvar > level) bl = bl->previous;
  if (bl) bl->upval = 1;
}


static void singlevaraux (FuncState *fs, TString *n, expdesc *var, int base) {
  if (fs == NULL)  /* no more levels? */
    init_exp(var, VGLOBAL, NO_REG);  /* default is global variable */
  else {
    int v = searchvar(fs, n);  /* look up at current level */
    if (v >= 0) {
      init_exp(var, VLOCAL, v);
      if (!base)
        markupval(fs, v);  /* local will be used as an upval */
    }
    else {  /* not found at current level; try upper one */
      singlevaraux(fs->prev, n, var, 0);
      if (var->k == VGLOBAL) {
        if (base)
          var->info = luaK_stringK(fs, n);  /* info points to global name */
      }
      else {  /* LOCAL or UPVAL */
        var->info = indexupvalue(fs, n, var);
        var->k = VUPVAL;  /* upvalue in this level */
      }
    }
  }
}


static TString *singlevar (LexState *ls, expdesc *var, int base) {
  TString *varname = str_checkname(ls);
  singlevaraux(ls->fs, varname, var, base);
  return varname;
}


static void adjust_assign (LexState *ls, int nvars, int nexps, expdesc *e) {
  FuncState *fs = ls->fs;
  int extra = nvars - nexps;
  if (hasmultret(e->k)) {
    extra++;  /* includes call itself */
    if (extra < 0) extra = 0;
    luaK_setreturns(fs, e, extra);  /* last exp. provides the difference */
    if (extra > 1) luaK_reserveregs(fs, extra-1);
  }
  else {
    if (e->k != VVOID) luaK_exp2nextreg(fs, e);  /* close last expression */
    if (extra > 0) {
      int reg = fs->freereg;
      luaK_reserveregs(fs, extra);
      luaK_nil(fs, reg, extra);
    }
  }
}


static void enterlevel (LexState *ls) {
  if (++ls->L->nCcalls > LUAI_MAXCCALLS)
	luaX_lexerror(ls, "chunk has too many syntax levels", 0);
}


#define leavelevel(ls)	((ls)->L->nCcalls--)


static void enterblock (FuncState *fs, BlockCnt *bl, int isbreakable) {
  bl->breaklist = NO_JUMP;
  bl->isbreakable = isbreakable;
  bl->nactvar = fs->nactvar;
  bl->upval = 0;
  bl->previous = fs->bl;
  fs->bl = bl;
  lua_assert(fs->freereg == fs->nactvar);
}


static void leaveblock (FuncState *fs) {
  BlockCnt *bl = fs->bl;
  fs->bl = bl->previous;
  removevars(fs->ls, bl->nactvar);
  if (bl->upval)
    luaK_codeABC(fs, OP_CLOSE, bl->nactvar, 0, 0);
  lua_assert(bl->nactvar == fs->nactvar);
  fs->freereg = fs->nactvar;  /* free registers */
  luaK_patchtohere(fs, bl->breaklist);
}


static void pushclosure (LexState *ls, FuncState *func, expdesc *v) {
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  int oldsize = f->sizep;
  int i;
  luaM_growvector(ls->L, f->p, fs->np, f->sizep, Proto *,
                  MAXARG_Bx, "constant table overflow");
  while (oldsize < f->sizep) f->p[oldsize++] = NULL;
  f->p[fs->np++] = func->f;
  luaC_objbarrier(ls->L, f, func->f);
  init_exp(v, VRELOCABLE, luaK_codeABx(fs, OP_CLOSURE, 0, fs->np-1));
  for (i=0; i<func->f->nups; i++) {
    OpCode o = (func->upvalues[i].k == VLOCAL) ? OP_MOVE : OP_GETUPVAL;
    luaK_codeABC(fs, o, 0, func->upvalues[i].info, 0);
  }
}


static void open_func (LexState *ls, FuncState *fs) {
  lua_State *L = ls->L;
  Proto *f = luaF_newproto(L);
  fs->f = f;
  fs->prev = ls->fs;  /* linked list of funcstates */
  fs->ls = ls;
  fs->L = L;
  ls->fs = fs;
  fs->pc = 0;
  fs->lasttarget = 0;
  fs->jpc = NO_JUMP;
  fs->freereg = 0;
  fs->nk = 0;
  fs->np = 0;
  fs->nlocvars = 0;
  fs->nactvar = 0;
  fs->nregvar = 0;
  fs->bl = NULL;
  f->source = ls->source;
  f->maxstacksize = 2;  /* registers 0/1 are always valid */
  fs->h = luaH_new(L, 0, 0);
  /* anchor table of constants and prototype (to avoid being collected) */
  sethvalue2s(L, L->top, fs->h);
  incr_top(L);
  setptvalue2s(L, L->top, f);
  incr_top(L);
}


static void close_func (LexState *ls) {
  lua_State *L = ls->L;
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  removevars(ls, 0);
  luaK_codeABC(fs, OP_RETURN, 0, 1, 0);  /* final return */
  luaM_reallocvector(L, f->code, f->sizecode, fs->pc, Instruction);
  f->sizecode = fs->pc;
  luaM_reallocvector(L, f->lineinfo, f->sizelineinfo, fs->pc, int);
  f->sizelineinfo = fs->pc;
  luaM_reallocvector(L, f->k, f->sizek, fs->nk, TValue);
  f->sizek = fs->nk;
  luaM_reallocvector(L, f->p, f->sizep, fs->np, Proto *);
  f->sizep = fs->np;
  luaM_reallocvector(L, f->locvars, f->sizelocvars, fs->nlocvars, LocVar);
  f->sizelocvars = fs->nlocvars;
  luaM_reallocvector(L, f->upvalues, f->sizeupvalues, f->nups, TString *);
  f->sizeupvalues = f->nups;
  lua_assert(luaG_checkcode(f));
  lua_assert(fs->bl == NULL);
  ls->fs = fs->prev;
  L->top -= 2;  /* remove table and prototype from the stack */
  /* last token read was anchored in defunct function; must reanchor it */
  if (fs) anchor_token(ls);
}


Proto *luaY_parser (lua_State *L, ZIO *z, Mbuffer *buff, Mbuffer *dbuff,
                    const char *name) {
  struct LexState lexstate;
  struct FuncState funcstate;
  lexstate.buff = buff;
  lexstate.dbuff = dbuff;
  luaX_setinput(L, &lexstate, z, luaS_new(L, name));
  open_func(&lexstate, &funcstate);
  funcstate.f->is_vararg = NEWSTYLEVARARG;
  next(&lexstate);  /* read first token */
  if (lexstate.t.token == TK_DOCSTRING) {
    /* Capture this string as the module's documentation */
    funcstate.f->documentation = lexstate.t.seminfo.ts;
    next(&lexstate);
  }
  chunk(&lexstate);
  check(&lexstate, TK_EOS);
  close_func(&lexstate);
  lua_assert(funcstate.prev == NULL);
  lua_assert(funcstate.f->nups == 0);
  lua_assert(lexstate.fs == NULL);
  return funcstate.f;
}



/*============================================================*/
/* GRAMMAR RULES */
/*============================================================*/


static void field (LexState *ls, expdesc *v) {
  /* field -> ['.' | ':'] NAME */
  FuncState *fs = ls->fs;
  expdesc key;
  luaK_exp2anyreg(fs, v);
  next(ls);  /* skip the dot or colon */
  checkname(ls, &key);
  luaK_indexed(fs, v, &key);
}


static void yindex (LexState *ls, expdesc *v) {
  /* index -> '[' expr ']' */
  next(ls);  /* skip the '[' */
  expr(ls, v);
  luaK_exp2val(ls->fs, v);
  checknext(ls, ']');
}


/*
** {======================================================================
** Rules for Constructors
** =======================================================================
*/


struct ConsControl {
  expdesc v;  /* last list item read */
  expdesc *t;  /* table descriptor */
  int nh;  /* total number of `record' elements */
  int na;  /* total number of array elements */
  int tostore;  /* number of array elements pending to be stored */
  int vecbase;  /* stack index of first list element */
  TString *doc;
};


static void recfield (LexState *ls, struct ConsControl *cc) {
  /* recfield -> (NAME | `['exp1`]') = exp1 */
  /* recfield -> function funcname body */
  FuncState *fs = ls->fs;
  int reg = ls->fs->freereg;
  expdesc key, val;
  if (ls->t.token == TK_NAME) {
    luaY_checklimit(fs, cc->nh, MAX_INT, "items in a constructor");
    checkname(ls, &key);
  }
  else if (ls->t.token == TK_FUNCTION) {
    int line = ls->linenumber;
    int needsself = 0;
    next(ls);
    if( ls->t.token == ':' ) {
      needsself = 1;
      next(ls);
    }
    cc->nh++;
    checkname(ls, &key);
    luaK_exp2RK(fs, &key);
    body(ls, &val, needsself, line);
    luaK_codeABC(fs, OP_SETTABLE, cc->t->info, luaK_exp2RK(fs, &key),
                                             luaK_exp2RK(fs, &val));
    fs->freereg = reg;  /* free registers */
    return;
  }
  else  /* ls->t.token == '[' */
    yindex(ls, &key);
  cc->nh++;
  checknext(ls, '=');
  luaK_exp2RK(fs, &key);
  expr(ls, &val);
  luaK_codeABC(fs, OP_SETTABLE, cc->t->info, luaK_exp2RK(fs, &key),
                                             luaK_exp2RK(fs, &val));
  fs->freereg = reg;  /* free registers */
}

/* multiforce: <0 forced >0 multi (and forced) */
static void closelistfield (FuncState *fs, struct ConsControl *cc, int multiforce) {
  int tostore = cc->tostore;
  if (cc->v.k == VVOID) {
    if (tostore == 0) return;  /* there is no list item */
  }
  else if (hasmultret(cc->v.k) && (multiforce > 0)) {
    luaK_setmultret(fs, &cc->v);
    tostore = 0;
  }
  else
    luaK_exp2nextreg(fs, &cc->v);
  if (cc->tostore == LFIELDS_PER_FLUSH || multiforce != 0) {
    luaK_codeABC(fs, OP_SETLIST, cc->t->info, tostore, cc->vecbase);
    fs->freereg = cc->vecbase;
    cc->tostore = 0;  /* no more items pending */
  }
  cc->v.k = VVOID;
}


static void listfield (LexState *ls, struct ConsControl *cc) {
  expr(ls, &cc->v);
  luaY_checklimit(ls->fs, cc->na, MAXARG_Bx, "items in a constructor");
  cc->na++;
  cc->tostore++;
}


static void forcomprehensionbody (LexState *ls, expdesc *t);

static void constructfields (LexState *ls, struct ConsControl *cc, int terminator) {
  FuncState *fs = ls->fs;
  cc->vecbase = fs->freereg;
  while (ls->t.token != terminator) {
    lua_assert(cc->v.k == VVOID || cc->tostore > 0);
    switch(ls->t.token) {
      case TK_FUNCTION:
      case TK_NAME: {  /* may be listfields or recfields */
        lookahead(ls);
        if (ls->lookahead.token != '=' && ls->t.token == TK_NAME)  /* expression? */
          listfield(ls, cc);
        else
          recfield(ls, cc);
        break;
      }
      case '[': {  /* constructor_item -> recfield */
        recfield(ls, cc);
        break;
      }
      case TK_FOR: {  /* The moment we've all been waiting for */
        closelistfield(fs, cc, -1);

        /* Interim code. The following saves (and then restores) the current
         * active register mapping, in order to allow the embedded forclause
         * to create locals (and even, I suppose, upvalues). This is not
         * particularly efficient, but fixing it properly requires a slight
         * redesign of the local variable lookup mechanism
         */
        {
          int i;
          int nactvar = fs->nactvar;
          int nregvar = fs->nregvar;
          int freereg = fs->freereg;
          unsigned short actvar[LUAI_MAXVARS];
          /*
           *  We need to create fake local variables for the part of the
           * stack we're skipping over. And we need to save the registered
           * but not yet active mapping.
           */
          for (i = nactvar; i < nregvar; ++i)
            actvar[i] = fs->actvar[i];
          for (i = nactvar; i < freereg; ++i)
            new_localvarliteral(ls, "(temp)", i - nactvar);
          adjustlocalvars(ls, freereg - nactvar);
          fs->nregvar = freereg;  /* We haven't preregistered anything */

          forclause(ls, ls->linenumber, 1, NULL, forcomprehensionbody, cc->t);

          /* Now undo the damage */
          removevars(ls, nactvar);
          fs->nregvar = nregvar;
          for (i = nactvar; i < nregvar; ++i)
            fs->actvar[i] = actvar[i];
        }
        break;
      }
      default: {  /* constructor_part -> listfield */
        listfield(ls, cc);
        break;
      }
    }
    if (testnext(ls, ';')) closelistfield(fs, cc, 1);
    else if (ls->t.token == terminator) break;
    else {
      testnext(ls, ',');
      closelistfield(fs, cc, 0);
    }
  }
  closelistfield(fs, cc, 1);
}


static void forcomprehensionbody (LexState *ls, expdesc *t) {
  struct ConsControl cc;
  cc.na = cc.nh = cc.tostore = 0;
  cc.t = t;
  init_exp(&cc.v, VVOID, 0); /* no value (yet) */

  if (ls->t.token == TK_NAME &&
      strcmp(getstr(ls->t.seminfo.ts), "yield") == 0) 
    next(ls);
  else
    luaX_syntaxerror(ls, 
        luaO_pushfstring(ls->L, LUA_QS " expected", "yield"));

  constructfields(ls, &cc, TK_END);
  /* 'END' will be checked on the way out, honest */
}


static void constructor (LexState *ls, expdesc *t) {
  /* constructor -> ?? */
  FuncState *fs = ls->fs;
  int line = ls->linenumber;
  int pc = luaK_codeABC(fs, OP_NEWTABLE, 0, 0, 0);
  struct ConsControl cc;
  int reg;
  cc.na = cc.nh = cc.tostore = 0;
  cc.t = t;
  cc.doc = NULL;
  init_exp(t, VRELOCABLE, pc);
  init_exp(&cc.v, VVOID, 0);  /* no value (yet) */
  luaK_exp2nextreg(ls->fs, t);  /* fix it at stack top (for gc) */
  reg = fs->freereg;
  luaK_reserveregs(fs, 1);  /* For table index */
  
  checknext(ls, '{');
  if (ls->t.token == TK_DOCSTRING) {
    /* Consume the documentation string and put it into the ConsControl */
    cc.doc = ls->t.seminfo.ts;
    next(ls);
  }
  constructfields(ls, &cc, '}');
  check_match(ls, '}', '{', line);
  SETARG_B(fs->f->code[pc], luaO_int2fb(cc.na)); /* set initial array size */
  SETARG_C(fs->f->code[pc], luaO_int2fb(cc.nh));  /* set initial table size */
  if (cc.doc != NULL) {
    /* We had a documentation string to process... */
    /* First up, anchor it in an expression... */
    expdesc dsexp;
    int base = ls->fs->freereg;
    codestring(ls, &dsexp, cc.doc);
    luaK_codeABC(fs, OP_TABDOC, cc.t->info, luaK_exp2anyreg(fs, &dsexp), 0);
    ls->fs->freereg = base;
  }
    
  ls->fs->freereg = reg; /* Ditch the table index */
}

/* }====================================================================== */



static void parlist (LexState *ls) {
  /* parlist -> [ param { `,' param } ] */
  FuncState *fs = ls->fs;
  Proto *f = fs->f;
  int nparams = 0;
  f->is_vararg = 0;
  if (ls->t.token != ')') {  /* is `parlist' not empty? */
    do {
      switch (ls->t.token) {
        case TK_NAME: {  /* param -> NAME */
          new_localvar(ls, str_checkname(ls), nparams++);
          break;
        }
        case TK_DOTS: {  /* param -> `...' */
          next(ls);
          /* use `arg' as default name */
          new_localvarliteral(ls, "arg", nparams++);
          f->is_vararg = 1;
          break;
        }
        default: luaX_syntaxerror(ls, "<name> or " LUA_QL("...") " expected");
      }
    } while (!f->is_vararg && testnext(ls, ','));
  }
  adjustlocalvars(ls, nparams);
  f->numparams = fs->nactvar - f->is_vararg;
  luaK_reserveregs(fs, fs->nactvar);  /* reserve register for parameters */
}


static void body (LexState *ls, expdesc *e, int needself, int line) {
  /* body ->  `(' parlist `)' chunk END */
  FuncState new_fs;
  open_func(ls, &new_fs);
  new_fs.f->linedefined = line;
  checknext(ls, '(');
  if (needself) {
    new_localvarliteral(ls, "self", 0);
    adjustlocalvars(ls, 1);
  }
  parlist(ls);
  checknext(ls, ')');
  if (ls->t.token == TK_DOCSTRING) {
    /* Capture this string as the function's documentation */
    new_fs.f->documentation = ls->t.seminfo.ts;
    next(ls);
  }
  chunk(ls);
  new_fs.f->lastlinedefined = ls->linenumber;
  check_match(ls, TK_END, TK_FUNCTION, line);
  if (needself) {
    /* Rename initial variable to (self) which can't be done from within
     * Lua source code itself. This serves as a sentinal for method functions
     */
    new_fs.f->locvars[0].varname = luaX_newstring(ls, "(self)", 6);
  }  
  close_func(ls);
  pushclosure(ls, &new_fs, e);
}


static int explist1 (LexState *ls, expdesc *v) {
  /* explist1 -> expr { `,' expr } */
  int n = 1;  /* at least one expression */
  expr(ls, v);
  while (testnext(ls, ',')) {
    luaK_exp2nextreg(ls->fs, v);
    expr(ls, v);
    n++;
  }
  return n;
}


static void funcargs (LexState *ls, expdesc *f) {
  FuncState *fs = ls->fs;
  expdesc args;
  int base, nparams;
  int line = ls->linenumber;
  switch (ls->t.token) {
    case '(': {  /* funcargs -> `(' [ explist1 ] `)' */
      if (line != ls->lastline)
        luaX_syntaxerror(ls,"ambiguous syntax (function call x new statement)");
      next(ls);
      if (ls->t.token == ')')  /* arg list is empty? */
        args.k = VVOID;
      else {
        explist1(ls, &args);
        luaK_setmultret(fs, &args);
      }
      check_match(ls, ')', '(', line);
      break;
    }
    case '{': {  /* funcargs -> constructor */
      constructor(ls, &args);
      break;
    }
    case TK_STRING: {  /* funcargs -> STRING */
      codestring(ls, &args, ls->t.seminfo.ts);
      next(ls);  /* must use `seminfo' before `next' */
      break;
    }
    default: {
      luaX_syntaxerror(ls, "function arguments expected");
      return;
    }
  }
  lua_assert(f->k == VNONRELOC);
  base = f->info;  /* base register for call */
  if (hasmultret(args.k))
    nparams = LUA_MULTRET;  /* open call */
  else {
    if (args.k != VVOID)
      luaK_exp2nextreg(fs, &args);  /* close last argument */
    nparams = fs->freereg - (base+1);
  }
  init_exp(f, VCALL, luaK_codeABC(fs, OP_CALL, base, nparams+1, 2));
  luaK_fixline(fs, line);
  fs->freereg = base+1;  /* call remove function and arguments and leaves
                            (unless changed) one result */
}




/*
** {======================================================================
** Expression parsing
** =======================================================================
*/


static void prefixexp (LexState *ls, expdesc *v) {
  /* prefixexp -> NAME | '(' expr ')' */
  switch (ls->t.token) {
    case '(': {
      int line = ls->linenumber;
      next(ls);
      expr(ls, v);
      check_match(ls, ')', '(', line);
      luaK_dischargevars(ls->fs, v);
      return;
    }
    case TK_NAME: {
      singlevar(ls, v, 1);
      return;
    }
    default: {
      luaX_syntaxerror(ls, "unexpected symbol");
      return;
    }
  }
}


static void primaryexp (LexState *ls, expdesc *v) {
  /* primaryexp ->
        prefixexp { `.' NAME | `[' exp `]' | `:' NAME funcargs | funcargs } */
  FuncState *fs = ls->fs;
  prefixexp(ls, v);
  for (;;) {
    switch (ls->t.token) {
      case '.': {  /* field */
        field(ls, v);
        break;
      }
      case '[': {  /* `[' exp1 `]' */
        expdesc key;
        luaK_exp2anyreg(fs, v);
        yindex(ls, &key);
        luaK_indexed(fs, v, &key);
        break;
      }
      case ':': {  /* `:' NAME funcargs */
        expdesc key;
        next(ls);
        checkname(ls, &key);
        luaK_self(fs, v, &key);
        funcargs(ls, v);
        break;
      }
      case '(': case TK_STRING: case '{': {  /* funcargs */
        luaK_exp2nextreg(fs, v);
        funcargs(ls, v);
        break;
      }
      default: return;
    }
  }
}


static void simpleexp (LexState *ls, expdesc *v) {
  /* simpleexp -> NUMBER | STRING | NIL | true | false | ... |
                  constructor | FUNCTION body | primaryexp */
  switch (ls->t.token) {
    case TK_NUMBER: {
      init_exp(v, VK, luaK_numberK(ls->fs, ls->t.seminfo.r));
      break;
    }
    case TK_STRING: {
      codestring(ls, v, ls->t.seminfo.ts);
      break;
    }
    case TK_NIL: {
      init_exp(v, VNIL, 0);
      break;
    }
    case TK_TRUE: {
      init_exp(v, VTRUE, 0);
      break;
    }
    case TK_FALSE: {
      init_exp(v, VFALSE, 0);
      break;
    }
    case TK_DOTS: {  /* vararg */
      FuncState *fs = ls->fs;
      check_condition(ls, fs->f->is_vararg,
                      "cannot use " LUA_QL("...") " outside a vararg function");
      fs->f->is_vararg = NEWSTYLEVARARG;
      init_exp(v, VVARARG, luaK_codeABC(fs, OP_VARARG, 0, 1, 0));
      break;
    }
    case '{': {  /* constructor */
      constructor(ls, v);
      return;
    }
    case TK_FUNCTION: {
      int needself = 0;
      next(ls);
      if (ls->t.token == ':') {
        next(ls);
        needself = 1;
      }
      body(ls, v, needself, ls->linenumber);
      return;
    }
    default: {
      primaryexp(ls, v);
      return;
    }
  }
  next(ls);
}


static UnOpr getunopr (int op) {
  switch (op) {
    case TK_NOT: return OPR_NOT;
    case '-': return OPR_MINUS;
    case '*': return OPR_SIZE;
    default: return OPR_NOUNOPR;
  }
}


static BinOpr getbinopr (int op) {
  switch (op) {
    case '+': return OPR_ADD;
    case '-': return OPR_SUB;
    case '*': return OPR_MULT;
    case '/': return OPR_DIV;
    case '%': return OPR_MOD;
    case '^': return OPR_POW;
    case TK_CONCAT: return OPR_CONCAT;
    case TK_NE: return OPR_NE;
    case TK_EQ: return OPR_EQ;
    case '<': return OPR_LT;
    case TK_LE: return OPR_LE;
    case '>': return OPR_GT;
    case TK_GE: return OPR_GE;
    case TK_AND: return OPR_AND;
    case TK_OR: return OPR_OR;
    default: return OPR_NOBINOPR;
  }
}


static const struct {
  lu_byte left;  /* left priority for each binary operator */
  lu_byte right; /* right priority */
} priority[] = {  /* ORDER OPR */
   {6, 6}, {6, 6}, {7, 7}, {7, 7}, {7, 7},  /* `+' `-' `/' `%' */
   {10, 9}, {5, 4},                 /* power and concat (right associative) */
   {3, 3}, {3, 3},                  /* equality and inequality */
   {3, 3}, {3, 3}, {3, 3}, {3, 3},  /* order */
   {2, 2}, {1, 1}                   /* logical (and/or) */
};

#define UNARY_PRIORITY	8  /* priority for unary operators */


/*
** subexpr -> (simpleexp | unop subexpr) { binop subexpr }
** where `binop' is any binary operator with a priority higher than `limit'
*/
static BinOpr subexpr (LexState *ls, expdesc *v, unsigned int limit) {
  BinOpr op;
  UnOpr uop;
  enterlevel(ls);
  uop = getunopr(ls->t.token);
  if (uop != OPR_NOUNOPR) {
    next(ls);
    subexpr(ls, v, UNARY_PRIORITY);
    luaK_prefix(ls->fs, uop, v);
  }
  else simpleexp(ls, v);
  /* expand while operators have priorities higher than `limit' */
  op = getbinopr(ls->t.token);
  while (op != OPR_NOBINOPR && priority[op].left > limit) {
    expdesc v2;
    BinOpr nextop;
    next(ls);
    luaK_infix(ls->fs, op, v);
    /* read sub-expression with higher priority */
    nextop = subexpr(ls, &v2, priority[op].right);
    luaK_posfix(ls->fs, op, v, &v2);
    op = nextop;
  }
  leavelevel(ls);
  return op;  /* return first untreated operator */
}


static void expr (LexState *ls, expdesc *v) {
  subexpr(ls, v, 0);
}

/* }==================================================================== */



/*
** {======================================================================
** Rules for Statements
** =======================================================================
*/


static int block_follow (int token) {
  switch (token) {
    case TK_ELSE: case TK_ELSEIF: case TK_END:
    case TK_UNTIL: case TK_EOS:
      return 1;
    default: return 0;
  }
}


static void block (LexState *ls) {
  /* block -> chunk */
  FuncState *fs = ls->fs;
  BlockCnt bl;
  enterblock(fs, &bl, 0);
  chunk(ls);
  lua_assert(bl.breaklist == NO_JUMP);
  leaveblock(fs);
}


/*
** structure to chain all variables in the left-hand side of an
** assignment
*/
struct LHS_assign {
  struct LHS_assign *prev;
  expdesc v;  /* variable (global, local, upvalue, or indexed) */
};


/*
** check whether, in an assignment to a local variable, the local variable
** is needed in a previous assignment (to a table). If so, save original
** local value in a safe place and use this safe copy in the previous
** assignment.
*/
static void check_conflict (LexState *ls, struct LHS_assign *lh, expdesc *v) {
  FuncState *fs = ls->fs;
  int extra = fs->freereg;  /* eventual position to save local variable */
  int conflict = 0;
  for (; lh; lh = lh->prev) {
    if (lh->v.k == VINDEXED) {
      if (lh->v.info == v->info) {  /* conflict? */
        conflict = 1;
        lh->v.info = extra;  /* previous assignment will use safe copy */
      }
      if (lh->v.aux == v->info) {  /* conflict? */
        conflict = 1;
        lh->v.aux = extra;  /* previous assignment will use safe copy */
      }
    }
  }
  if (conflict) {
    luaK_codeABC(fs, OP_MOVE, fs->freereg, v->info, 0);  /* make copy */
    luaK_reserveregs(fs, 1);
  }
}


static void assignment (LexState *ls, struct LHS_assign *lh, int nvars) {
  expdesc e;
  check_condition(ls, VLOCAL <= lh->v.k && lh->v.k <= VINDEXED,
                      "syntax error");
  if (testnext(ls, ',')) {  /* assignment -> `,' primaryexp assignment */
    struct LHS_assign nv;
    nv.prev = lh;
    primaryexp(ls, &nv.v);
    if (nv.v.k == VLOCAL)
      check_conflict(ls, lh, &nv.v);
    assignment(ls, &nv, nvars+1);
  }
  else {  /* assignment -> `=' explist1 */
    int nexps;
    checknext(ls, '=');
    nexps = explist1(ls, &e);
    if (nexps != nvars) {
      adjust_assign(ls, nvars, nexps, &e);
      if (nexps > nvars)
        ls->fs->freereg -= nexps - nvars;  /* remove extra values */
    }
    else {
      luaK_setoneret(ls->fs, &e);  /* close last expression */
      luaK_storevar(ls->fs, &lh->v, &e);
      return;  /* avoid default */
    }
  }
  init_exp(&e, VNONRELOC, ls->fs->freereg-1);  /* default assignment */
  luaK_storevar(ls->fs, &lh->v, &e);
}


static int cond (LexState *ls) {
  /* cond -> exp */
  expdesc v;
  expr(ls, &v);  /* read condition */
  if (v.k == VNIL) v.k = VFALSE;  /* `falses' are all equal here */
  luaK_goiftrue(ls->fs, &v);
  luaK_patchtohere(ls->fs, v.t);
  return v.f;
}


/*
** The while statement optimizes its code by coding the condition
** after its body (and thus avoiding one jump in the loop).
*/


/*
** the call `luaK_goiffalse' may grow the size of an expression by
** at most this:
*/
#define EXTRAEXP	5

static void whilestat (LexState *ls, int line) {
  /* whilestat -> WHILE cond DO block END */
  Instruction codeexp[LUAI_MAXEXPWHILE + EXTRAEXP];
  int lineexp;
  int i;
  int sizeexp;
  FuncState *fs = ls->fs;
  int whileinit, blockinit, expinit;
  expdesc v;
  BlockCnt bl;
  next(ls);  /* skip WHILE */
  whileinit = luaK_jump(fs);  /* jump to condition (which will be moved) */
  expinit = luaK_getlabel(fs);
  expr(ls, &v);  /* parse condition */
  if (v.k == VK) v.k = VTRUE;  /* `trues' are all equal here */
  lineexp = ls->linenumber;
  luaK_goiffalse(fs, &v);
  luaK_concat(fs, &v.f, fs->jpc);
  fs->jpc = NO_JUMP;
  sizeexp = fs->pc - expinit;  /* size of expression code */
  if (sizeexp > LUAI_MAXEXPWHILE) 
    luaX_syntaxerror(ls, LUA_QL("while") " condition too complex");
  for (i = 0; i < sizeexp; i++)  /* save `exp' code */
    codeexp[i] = fs->f->code[expinit + i];
  fs->pc = expinit;  /* remove `exp' code */
  enterblock(fs, &bl, 1);
  checknext(ls, TK_DO);
  blockinit = luaK_getlabel(fs);
  block(ls);
  luaK_patchtohere(fs, whileinit);  /* initial jump jumps to here */
  /* move `exp' back to code */
  if (v.t != NO_JUMP) v.t += fs->pc - expinit;
  if (v.f != NO_JUMP) v.f += fs->pc - expinit;
  for (i=0; i<sizeexp; i++)
    luaK_code(fs, codeexp[i], lineexp);
  check_match(ls, TK_END, TK_WHILE, line);
  leaveblock(fs);
  luaK_patchlist(fs, v.t, blockinit);  /* true conditions go back to loop */
  luaK_patchtohere(fs, v.f);  /* false conditions finish the loop */
}


static void repeatstat (LexState *ls, int line) {
  /* repeatstat -> REPEAT block UNTIL cond */
  FuncState *fs = ls->fs;
  int repeat_init = luaK_getlabel(fs);
  int flist;
  BlockCnt bl;
  enterblock(fs, &bl, 1);
  next(ls);
  block(ls);
  check_match(ls, TK_UNTIL, TK_REPEAT, line);
  flist = cond(ls);
  luaK_patchlist(fs, flist, repeat_init);
  leaveblock(fs);
}


static int exp1 (LexState *ls) {
  expdesc e;
  int k;
  expr(ls, &e);
  k = e.k;
  luaK_exp2nextreg(ls->fs, &e);
  return k;
}


/* Conditional break (if necessary, insert an OP_CLOSE) */
static void condbreak (LexState *ls) {
  /* for... while cond -> exp */
  FuncState *fs = ls->fs;
  BlockCnt *bl = fs->bl;
  int upval = 0;
  expdesc v;
  expr(ls, &v);  /* read condition */
  /* Copied from breakstat */
  while (!bl->isbreakable) {
    upval |= bl->upval;
    bl = bl->previous;
    lua_assert(bl);
  }
  if (upval) {
    /* If the result is false, we need to insert a close */
    if (v.k == VK) v.k = VTRUE;   /* all trues look alike */
    luaK_goiffalse(ls->fs, &v);
    luaK_patchtohere(ls->fs, v.f);
    luaK_codeABC(fs, OP_CLOSE, bl->nactvar, 0, 0);
    luaK_concat(fs, &bl->breaklist, luaK_jump(fs));
    luaK_patchtohere(ls->fs, v.t);
  }
  else {
    /* Quite a bit simpler in the normal case */
  if (v.k == VNIL) v.k = VFALSE;  /* `falses' are all equal here */
  luaK_goiftrue(ls->fs, &v);
  luaK_patchtohere(ls->fs, v.t);
  luaK_concat(fs, &bl->breaklist, v.f);
  }
}


static void forinvertedloop (LexState *ls, int base, int nvars, int isnum,
                             struct LockstepControl *ctl) {
  FuncState *fs = ls->fs;
  BlockCnt *bl = fs->bl;
  int upval = 0;
  lua_assert(ctl);
  while (!bl->isbreakable) {
    upval |= bl->upval;
    bl = bl->previous;
    lua_assert(bl);
  }
  if (isnum) {
    ctl->loop = ctl->bodyj = luaK_codeAsBx(fs, OP_FORLOOP, base, NO_JUMP);
  }
  else {
    ctl->loop = luaK_codeABC(fs, OP_TFORLOOP, base, 0, nvars);
    ctl->bodyj = luaK_jump(fs);
  }
  if (upval)
    luaK_codeABC(fs, OP_CLOSE, bl->nactvar, 0, 0);
  luaK_concat(fs, &bl->breaklist, luaK_jump(fs));
  luaK_patchtohere(fs, ctl->prepj);
}


/* OK, call me stubborn. This was the only way I could think of to get
 * andfor to work.
 *
 * This function scans a range of opcodes, and corrects all register
 * references >= base by adding offset to them. Since we can't easily
 * tell whether A refers to a register or not, base must be at least
 * 2 (which will avoid JMP (0) and LE,LT,EQ (0,1) )
 */

static void relocate (FuncState *fs, int frompc, int topc, int base, int offset) {
  int pc;
  Instruction *code = fs->f->code;
  int hwm = base;
  for (pc = frompc; pc < topc; ++pc) {
    Instruction i = code[pc];
    OpCode op = GET_OPCODE(i);
    int a = GETARG_A(i);
    int b = GETARG_B(i);
    int c = GETARG_C(i);
    /* Loosely grabbed from ldebug.c, special cases for hwm */
    /* Many of these tests are probably not strictly necessary, since the
     * registers will be used in some opcode anyway. But better safe than
     * sorry
     */
    switch(op) {
      case OP_SELF:
      case OP_NEWTABLE:
        if (hwm < a + 1) hwm = a + 1; 
        break;
      case OP_TFORLOOP:
        if (hwm < a + 2 + c) hwm = a + 2 + c;
        if (hwm < b - 1 + c) hwm = b - 1 + c;
        break;
      case OP_FORLOOP:
      case OP_FORPREP:
        if (hwm < a + 3) hwm = a + 3;
        break;
      case OP_CALL:
      case OP_TAILCALL:
        if (hwm < a + b - 1) hwm = a + b - 1;
        if (hwm < a + c - 1) hwm = a + c - 1;
        break;
      case OP_RETURN:
        if (hwm < a + b - 1) hwm = a + b - 1;
        break;
      case OP_SETLIST:
        if (hwm < c + b + 1) hwm = c + b + 1;
        break;
      case OP_VARARG:
        if (hwm < a + b - 2) hwm = a + b -2;
        break;
      default:
        break;
    }
    if (a >= base) {
      if (a > hwm) hwm = a;
      SETARG_A(i, a + offset);
    }
    if (getOpMode(op) == iABC) {
      if (getBMode(op) >= OpArgR) {
        if (!ISK(b) && b >= base) {
          if (b > hwm) hwm = b;
          SETARG_B(i, b + offset);
        }
      }
      if (getCMode(op) >= OpArgR) {
        if (!ISK(c) && c >= base) {
          if (c > hwm) hwm = c;
          SETARG_C(i, c + offset);
        }
      }
    }
    code[pc] = i;
  }
  luaK_checkstack(fs, hwm + offset - fs->freereg);
}


static void forbody (LexState *ls, int base, int line, int nvars, int isnum,
                     struct LockstepControl *ctl,
                     void(*bodyfunc)(LexState *ls, expdesc *t), expdesc *t) {
  /* forbody -> DO block */
  BlockCnt bl;
  FuncState *fs = ls->fs;
  int prep, ourbody;
  int continuelist = NO_JUMP;
  int tempbase = 0;
  int frompc, topc, regbase;
  adjustlocalvars(ls, 3);  /* control variables */
  prep = isnum ? luaK_codeAsBx(fs, OP_FORPREP, base, NO_JUMP) : luaK_jump(fs);
  ourbody = prep + 1;
  if (ctl) {
    /* Info from forinvertedloop */
    luaK_patchtohere(fs, ctl->bodyj);
    /* Info back to caller */
    ctl->prepj = prep;
  }
  enterblock(fs, &bl, 0);  /* scope for declared variables */
  adjustlocalvars(ls, nvars);
  luaK_reserveregs(fs, nvars);
  /* Now we can output any tests we run into */
  frompc = fs->pc;
  regbase = fs->freereg;
  for (;;) {
    if (ls->t.token == TK_NAME &&
        strcmp(getstr(ls->t.seminfo.ts), "where") == 0) {
      next(ls);
      if (ctl)
        luaK_patchlist(fs, cond(ls), ctl->loop);
      else
        luaK_concat(fs, &continuelist, cond(ls));
    }
    else if (ls->t.token == TK_WHILE) {
      next(ls);
      condbreak(ls);
    }
    else break;
  }
  if (ls->t.token == TK_NAME &&
           strcmp(getstr(ls->t.seminfo.ts), "andfor") == 0) {
    struct LockstepControl theirctl;
    theirctl.prepj = prep;
    if (!ctl || ctl->maxnvars < nvars)
      theirctl.maxnvars = nvars;
    else
      theirctl.maxnvars = ctl->maxnvars;
    topc = fs->pc;
    
    forclause(ls, line, 0, &theirctl, bodyfunc, t);
    
    prep = theirctl.prepj;
    tempbase = theirctl.tempbase;

    /* Any use of temporaries in while/where clauses (if any) will
     * overwrite downstream for variables. So we relocate the code
     * if we have to.
     */
    relocate(fs, frompc, topc, regbase, tempbase - regbase);

    /* If we're a TFORLOOP and we've already issued the TFORLOOP in
     * forinverted control, we need to change its B value to tempbase.
     */
    if (ctl && !isnum) 
      SETARG_B(fs->f->code[ctl->loop], tempbase);
  }
  else {
    if (ctl) {
      /* *This* for is a lockstep for, but the one inside (or the body)
       * isn't. If there are TFORLOOPs back up the chain, we need to
       * make sure they have room for their temporary space, and
       * tell them where it is.
       */
      ctl->tempbase = fs->freereg;  /* This might be one too high */
      luaK_checkstack(fs, ctl->maxnvars);
    }
    while (ls->t.token == TK_NAME &&
           strcmp(getstr(ls->t.seminfo.ts), "with") == 0) {
      next(ls); 
      localstat(ls);
    }
    if (ls->t.token == TK_FOR)
      forclause(ls, line, 0, NULL, bodyfunc, t);
    else
      bodyfunc(ls, t);
  }
  leaveblock(fs);  /* end of scope for declared variables */
  if (!ctl) {
    int endfor;
    luaK_patchtohere(fs, continuelist);
    luaK_patchtohere(fs, prep);
    endfor = (isnum) ? luaK_codeAsBx(fs, OP_FORLOOP, base, NO_JUMP) :
                       luaK_codeABC(fs, OP_TFORLOOP, base, tempbase, nvars);
    luaK_fixline(fs, line);  /* pretend that `OP_FOR' starts the loop */
    luaK_patchlist(fs, (isnum ? endfor : luaK_jump(fs)), ourbody);
  }
}


static void fornum (LexState *ls, TString *varname, int line, struct LockstepControl *ctl,
                    void(*bodyfunc)(LexState *ls, expdesc *t), expdesc *t) {
  /* fornum -> NAME = exp1,exp1[,exp1] forbody */
  FuncState *fs = ls->fs;
  int base = fs->freereg;
  new_localvarliteral(ls, "(for index)", 0);
  new_localvarliteral(ls, "(for limit)", 1);
  new_localvarliteral(ls, "(for step)", 2);
  new_localvar(ls, varname, 3);
  checknext(ls, '=');
  if (ctl)
    forinvertedloop(ls, base, 1 /* ignored */, 1, ctl);
  exp1(ls);  /* initial value */
  checknext(ls, ',');
  exp1(ls);  /* limit */
  if (testnext(ls, ','))
    exp1(ls);  /* optional step */
  else {  /* default step = 1 */
    luaK_codeABx(fs, OP_LOADK, fs->freereg, luaK_numberK(fs, 1));
    luaK_reserveregs(fs, 1);
  }
  forbody(ls, base, line, 1, 1, ctl, bodyfunc, t);
}


static void forlist (LexState *ls, TString *indexname, struct LockstepControl *ctl,
                     void(*bodyfunc)(LexState *ls, expdesc *t), expdesc *t) {
  /* forlist -> NAME {,NAME} IN explist1 forbody */
  FuncState *fs = ls->fs;
  expdesc e;
  int nvars = 0;
  int line;
  int base = fs->freereg;
  /* create control variables */
  new_localvarliteral(ls, "(for generator)", nvars++);
  new_localvarliteral(ls, "(for state)", nvars++);
  new_localvarliteral(ls, "(for control)", nvars++);
  /* create declared variables */
  new_localvar(ls, indexname, nvars++);
  while (testnext(ls, ','))
    new_localvar(ls, str_checkname(ls), nvars++);
  checknext(ls, TK_IN);
  line = ls->linenumber;
  if (ctl) 
    forinvertedloop(ls, base, nvars - 3, 0, ctl);
  adjust_assign(ls, 3, explist1(ls, &e), &e);
  luaK_checkstack(fs, 3);  /* extra space to call generator */
  forbody(ls, base, line, nvars - 3, 0, ctl, bodyfunc, t);
}

static void forclause (LexState *ls, int line, int breakable, struct LockstepControl *ctl,
                     void(*bodyfunc)(LexState *ls, expdesc *t), expdesc *t) {
  /* forstat -> FOR (fornum | forlist) END */
  FuncState *fs = ls->fs;
  TString *varname;
  BlockCnt bl;
  enterblock(fs, &bl, breakable);  /* scope for loop and control variables */
  next(ls);  /* skip `for' */
  varname = str_checkname(ls);  /* first variable name */
  switch (ls->t.token) {
    case '=': fornum(ls, varname, line, ctl, bodyfunc, t); break;
    case ',': case TK_IN: forlist(ls, varname, ctl, bodyfunc, t); break;
    default: luaX_syntaxerror(ls, LUA_QL("=") " or " LUA_QL("in") " expected");
  }
  if (breakable)
    check_match(ls, TK_END, TK_FOR, line);
  leaveblock(fs);  /* loop scope (`break' jumps to this point) */
}


static void forstmtbody (LexState *ls, expdesc *unused) {
  checknext(ls, TK_DO);
  block(ls);
}


static void forstat (LexState *ls, int line) {
  forclause(ls, line, 1, NULL, forstmtbody, NULL);
}


static int test_then_block (LexState *ls) {
  /* test_then_block -> [IF | ELSEIF] cond THEN block */
  int flist;
  next(ls);  /* skip IF or ELSEIF */
  flist = cond(ls);
  checknext(ls, TK_THEN);
  block(ls);  /* `then' part */
  return flist;
}


static void ifstat (LexState *ls, int line) {
  /* ifstat -> IF cond THEN block {ELSEIF cond THEN block} [ELSE block] END */
  FuncState *fs = ls->fs;
  int flist;
  int escapelist = NO_JUMP;
  flist = test_then_block(ls);  /* IF cond THEN block */
  while (ls->t.token == TK_ELSEIF) {
    luaK_concat(fs, &escapelist, luaK_jump(fs));
    luaK_patchtohere(fs, flist);
    flist = test_then_block(ls);  /* ELSEIF cond THEN block */
  }
  if (ls->t.token == TK_ELSE) {
    luaK_concat(fs, &escapelist, luaK_jump(fs));
    luaK_patchtohere(fs, flist);
    next(ls);  /* skip ELSE (after patch, for correct line info) */
    block(ls);  /* `else' part */
  }
  else
    luaK_concat(fs, &escapelist, flist);
  luaK_patchtohere(fs, escapelist);
  check_match(ls, TK_END, TK_IF, line);
}


static void localfunc (LexState *ls) {
  expdesc v, b;
  FuncState *fs = ls->fs;
  new_localvar(ls, str_checkname(ls), 0);
  init_exp(&v, VLOCAL, fs->freereg);
  luaK_reserveregs(fs, 1);
  adjustlocalvars(ls, 1);
  body(ls, &b, 0, ls->linenumber);
  luaK_storevar(fs, &v, &b);
  /* debug information will only see the variable after this point! */
  getlocvar(fs, fs->nactvar - 1).startpc = fs->pc;
}


static void localstat (LexState *ls) {
  /* stat -> LOCAL NAME {`,' NAME} [`=' explist1] */
  int nvars = 0;
  int nexps;
  expdesc e;
  do {
    new_localvar(ls, str_checkname(ls), nvars++);
  } while (testnext(ls, ','));
  if (testnext(ls, '='))
    nexps = explist1(ls, &e);
  else {
    e.k = VVOID;
    nexps = 0;
  }
  adjust_assign(ls, nvars, nexps, &e);
  adjustlocalvars(ls, nvars);
}


static int funcname (LexState *ls, expdesc *v) {
  /* funcname -> NAME {field} [`:' NAME] */
  int needself = 0;
  singlevar(ls, v, 1);
  while (ls->t.token == '.')
    field(ls, v);
  if (ls->t.token == ':') {
    needself = 1;
    field(ls, v);
  }
  return needself;
}


static void funcstat (LexState *ls, int line) {
  /* funcstat -> FUNCTION funcname body */
  int needself = 0;
  expdesc v, b;
  next(ls);  /* skip FUNCTION */
  if (ls->t.token == ':') {
    needself = 1;
    next(ls);
  }
  needself = needself | funcname(ls, &v);
  body(ls, &b, needself, line);
  luaK_storevar(ls->fs, &v, &b);
  luaK_fixline(ls->fs, line);  /* definition `happens' in the first line */
}


static void exprstat (LexState *ls) {
  /* stat -> func | assignment */
  FuncState *fs = ls->fs;
  struct LHS_assign v;
  primaryexp(ls, &v.v);
  if (v.v.k == VCALL)  /* stat -> func */
    SETARG_C(getcode(fs, &v.v), 1);  /* call statement uses no results */
  else {  /* stat -> assignment */
    v.prev = NULL;
    assignment(ls, &v, 1);
  }
}


static void retstat (LexState *ls) {
  /* stat -> RETURN explist */
  FuncState *fs = ls->fs;
  expdesc e;
  int first, nret;  /* registers with returned values */
  next(ls);  /* skip RETURN */
  if (block_follow(ls->t.token) || ls->t.token == ';')
    first = nret = 0;  /* return no values */
  else {
    nret = explist1(ls, &e);  /* optional return values */
    if (hasmultret(e.k)) {
      luaK_setmultret(fs, &e);
      if (e.k == VCALL && nret == 1) {  /* tail call? */
        SET_OPCODE(getcode(fs,&e), OP_TAILCALL);
        lua_assert(GETARG_A(getcode(fs,&e)) == fs->nactvar);
      }
      first = fs->nactvar;
      nret = LUA_MULTRET;  /* return all values */
    }
    else {
      if (nret == 1)  /* only one single value? */
        first = luaK_exp2anyreg(fs, &e);
      else {
        luaK_exp2nextreg(fs, &e);  /* values must go to the `stack' */
        first = fs->nactvar;  /* return all `active' values */
        lua_assert(nret == fs->freereg - first);
      }
    }
  }
  luaK_codeABC(fs, OP_RETURN, first, nret+1, 0);
}


static void breakstat (LexState *ls) {
  /* stat -> BREAK */
  FuncState *fs = ls->fs;
  BlockCnt *bl = fs->bl;
  int upval = 0;
  next(ls);  /* skip BREAK */
  while (bl && !bl->isbreakable) {
    upval |= bl->upval;
    bl = bl->previous;
  }
  if (!bl)
    luaX_syntaxerror(ls, "no loop to break");
  if (upval)
    luaK_codeABC(fs, OP_CLOSE, bl->nactvar, 0, 0);
  luaK_concat(fs, &bl->breaklist, luaK_jump(fs));
}

/* We're borrowing these from lauxlib.c */

typedef struct DivertedLoadS {
  const char *s;
  size_t size;
} DivertedLoadS;


static const char *divertedgetS (lua_State *L, void *ud, size_t *size) {
  DivertedLoadS *ls = (DivertedLoadS *)ud;
  (void)L;
  if (ls->size == 0) return NULL;
  *size = ls->size;
  ls->size = 0;
  return ls->s;
}

/* This delimits the expressions in diverted strings. It should be
 * unlikely to appear in HTML or SQL since those are the most
 * common languages which will use it
 *
 * Since | is the current chosen delimiter there is the added cuteness
 * that two of DIVERTED_EXPRESSION_DELIMITER in a row will be considered
 * as part of the string rather than a reason to switch to expression mode.
 *
 * Also, DIVERTED_EXPRESSION_FORMAT_DELIMITER is used between |expr and | to
 * allow you to control the format of the value. E.g. |pi#1.4f|
 *
 * That optimisation is not passed on in the decision to use vdprint or not
 * so you still get the slightly slower __divertedstring call even if a
 * call to __verbatimdivertedstring would have done the job.
 */
#define DIVERTED_EXPRESSION_DELIMITER '|'
#define DIVERTED_EXPRESSION_FORMAT_DELIMITER '#'

#define DIVERTED_EXPRESSION_ACCUMULATE_CHAR(L,b,c) \
        if (b->n + 1 > b->buffsize) { \
          size_t newsize; \
          if (b->buffsize >= MAX_SIZET/2) \
            luaX_lexerror(ls, "lexical element too long", 0); \
          newsize = (b->buffsize * 2)+1; \
          luaZ_resizebuffer(L, b, newsize); \
        } \
        b->buffer[b->n++] = cast(char, c)


static void divertedstringstat (LexState *ls) {
  /* stat -> divertedstring */
  int f = 0, i = 0;
  const char* ts = getstr(ls->t.seminfo.ts);
  
  if( ls->t.seminfo.ts->tsv.len == 0 ) { 
    /* String of zero bytes is useless to us */
    next(ls);
    return;
  }
  /* We can't be sure a string is NULL terminated; so... */
  for (i = 0; i < ls->t.seminfo.ts->tsv.len; ++i, ++ts)
    if (*ts == DIVERTED_EXPRESSION_DELIMITER) { f = 1; break; }
  if( f == 0 ) {
    /* Basic case: verbatim string with no interpolation; process simply */
    int base;
    expdesc func, text;
    int line = ls->linenumber;
    ls->lastline = line;
    /* Remember the registers */
    base = ls->fs->freereg;
    /* 1 load __divertedstring into a register */
    singlevaraux(ls->fs, luaS_newliteral(ls->L, "__verbatimdivertedstring"), &func, 1);
    luaK_exp2nextreg(ls->fs, &func);
    /* 2 load the diverted string itself into a register */
    codestring(ls, &text, ls->t.seminfo.ts);
    luaK_exp2nextreg(ls->fs, &text);
    /* 3 make the function call, 1 arg, no returns */
    init_exp( &func, VCALL, luaK_codeABC(ls->fs, OP_CALL, base, 2, 1));
    luaK_fixline(ls->fs, line);
    ls->fs->freereg = base;
    /* 4. Move on. */
    next(ls);
    return;
  }
  /* More complex case of non-verbatim string needs a little more work
   * because we essentially have to construct a new lexer using a ZIO over
   * the TString of ls->t.seminfo.ts and get on with running it
   * over our string, We parse carefully until we hit a | and then we do
   * something odd with parsing an expression into a register somehow
   */
  {
    LexState newls;
    TString *newname;
    char buffer[2];
    int bufflen = 0;
    ZIO newz;
    DivertedLoadS str;
    int base = 0;
    expdesc func, temp;
    int line = ls->linenumber;
    int accumargs = 0;
    int regforformatstring;
    
    ls->lastline = line;
    
    {
      char *namebuff;
      namebuff = luaM_malloc(ls->L, 128+ls->source->tsv.len);
      snprintf(namebuff, 128, "Diverted string around line %d of ", 
               ls->linenumber);
      namebuff[strlen(namebuff)+ls->source->tsv.len] = 0;
      memcpy(namebuff+strlen(namebuff), getstr(ls->source),
             ls->source->tsv.len);
      newname = luaS_new(ls->L, namebuff);
      luaM_freemem(ls->L, namebuff, 128+ls->source->tsv.len);
    }
    /* We now have a name for our new lexer... 
     *
     * First up, we need to preserve any contents in the buffer which we
     * are about to lend to our new lex state...
     *
     * Since we're being called due to a TK_DIVERTEDSTRING, there should
     * be *AT MOST* two characters in the buffer, if there are more then
     * we're fucked...
     *
     */
    if (luaZ_bufflen(ls->buff) > 2) {
      /* Throw an error because we cannot compile this */
      luaX_syntaxerror(ls,
                       luaO_pushfstring(ls->L, 
                                        "Buffer too large when parsing "\
                                        "diverted string. Expected 0,1,2 "\
                                        
                                        "chars, got %d instead.", 
                                        luaZ_bufflen(ls->buff)));
    }
    /* Okay, we can compile this, copy the buffer length and chars into
     * our temporary storage...
     */
    bufflen = luaZ_bufflen(ls->buff);
    if (bufflen > 0) {
      buffer[0] = luaZ_buffer(ls->buff)[0];
      if (bufflen > 1)
        buffer[1] = luaZ_buffer(ls->buff)[1];
    }
    luaZ_bufflen(ls->buff) = 0;
    luaZ_bufflen(ls->dbuff) = 0; /* Empty out the dbuffer */
    /* Now construct the ZIO over our diverted string */
    str.s = getstr(ls->t.seminfo.ts);
    str.size = ls->t.seminfo.ts->tsv.len;
    luaZ_init(ls->L, &newz, &divertedgetS, &str);
    /* Feed in the buffer we're using */
    newls.buff = ls->buff;
    newls.dbuff = ls->dbuff;
    /* And now the lexer over this new ZIO */
    luaX_setinput(ls->L, &newls, &newz, newname);
    /* Copy in the function state */
    newls.fs = ls->fs;
    /* Remember the registers */
    base = ls->fs->freereg;
    /* load __divertedstring into a register */
    singlevaraux(ls->fs, luaS_newliteral(ls->L, "__divertedstring"), &func, 1);
    luaK_exp2nextreg(ls->fs, &func);
    
    regforformatstring = ls->fs->freereg;
    luaK_reserveregs(ls->fs, 1); /* Reserve one register for the format str */
    
    luaZ_resetbuffer(newls.buff); /* We accumulate non-expressions here */
    while (newls.t.token != TK_EOS && newls.current != EOZ) {
      /* We're not going to have a lookahead because that's rude
       * Simply read the next byte and accumulate if not a pipe
       */
      if (newls.current != DIVERTED_EXPRESSION_DELIMITER ||
          (newls.current == DIVERTED_EXPRESSION_DELIMITER &&
           luaZ_lookahead(newls.z) == DIVERTED_EXPRESSION_DELIMITER)) {
        /* We want to accumulate this into the buffer, this means
         * we have to resize the buffer if it's too small
         * This is save() from llex.c because it's static
         */
        DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                            newls.current);
        if (newls.current == DIVERTED_EXPRESSION_DELIMITER) {
          newls.current = zgetc(newls.z); /* next(newls) from llex.c */
        }
        if (newls.current == '%') {
          DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                              newls.current);
        }
        newls.current = zgetc(newls.z); /* next(newls) from llex.c */
      } else {
        /* Consume the pipe */
        newls.current = zgetc(newls.z); /* next(newls) from llex.c */
        newls.t.token = luaX_lex(&newls, &newls.t.seminfo);
        expr(&newls, &temp);
        luaK_exp2nextreg(ls->fs, &temp);
        accumargs++;
        /* Emit a % to the dbuffer */
        DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                            '%');
        /* The terminating pipe should have been absorbed into t.token */
        if (newls.t.token == DIVERTED_EXPRESSION_DELIMITER) {
          /* Okay, we haven't got a format string, emit an 's' */
          DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff, 
                                              's');
        } else {
          if (newls.t.token == DIVERTED_EXPRESSION_FORMAT_DELIMITER) {
            /* Righty, we got what we expected, transfer until delimiter */
            while (newls.current != DIVERTED_EXPRESSION_DELIMITER) {
              DIVERTED_EXPRESSION_ACCUMULATE_CHAR(newls.L, newls.dbuff,
                                                  newls.current);
              newls.current = zgetc(newls.z); /* next(ls) from llex.c */
            }
            newls.current = zgetc(newls.z); /* Consume the delimiter */
          } else {
            /* Dunno what it is, abort abort abort */
            luaX_syntaxerror(ls,
                             luaO_pushfstring(ls->L, 
                                              "Expected `%c' or `%c', "\
                                              "got `%s'  "\
                                              "while parsing " \
                                              "diverted string.",
                                              DIVERTED_EXPRESSION_DELIMITER,
                                              DIVERTED_EXPRESSION_FORMAT_DELIMITER,
                                              luaX_token2str(&newls,
                                                             newls.t.token)
                                              ));
          }
        }
      }
    }
    {
      int regof;
      codestring(ls, &temp, luaX_newstring(&newls, newls.dbuff->buffer,
                                           newls.dbuff->n));
      regof = ls->fs->freereg;
      luaK_exp2nextreg(ls->fs, &temp);
      luaZ_resetbuffer(newls.dbuff);
      luaK_codeABC(ls->fs, OP_MOVE, regforformatstring, regof, 0);
      ls->fs->freereg = regof;
    }
    /* Finally, code up the VCALL */
    init_exp( &func, VCALL, luaK_codeABC(ls->fs, OP_CALL, base, accumargs+2, 1));
    luaK_fixline(ls->fs, line);
    ls->fs->freereg = base;
    /* We restore the content of our old lex buffer */
    if (luaZ_sizebuffer(ls->buff) < LUA_MINBUFFER)
      luaZ_resizebuffer(ls->L, ls->buff, LUA_MINBUFFER /* at least 2 */);
    luaZ_bufflen(ls->buff) = bufflen;
    if (bufflen > 0) {
      luaZ_buffer(ls->buff)[0] = buffer[0];
      if (bufflen > 1)
        luaZ_buffer(ls->buff)[1] = buffer[1];
    }
    /* Finally, consume the lexeme from our primary lexer */
    next(ls);
  }
}

static int statement (LexState *ls) {
  int line = ls->linenumber;  /* may be needed for error messages */
  switch (ls->t.token) {
    case TK_IF: {  /* stat -> ifstat */
      ifstat(ls, line);
      return 0;
    }
    case TK_WHILE: {  /* stat -> whilestat */
      whilestat(ls, line);
      return 0;
    }
    case TK_DO: {  /* stat -> DO block END */
      next(ls);  /* skip DO */
      block(ls);
      check_match(ls, TK_END, TK_DO, line);
      return 0;
    }
    case TK_FOR: {  /* stat -> forstat */
      forstat(ls, line);
      return 0;
    }
    case TK_REPEAT: {  /* stat -> repeatstat */
      repeatstat(ls, line);
      return 0;
    }
    case TK_FUNCTION: {
      funcstat(ls, line);  /* stat -> funcstat */
      return 0;
    }
    case TK_LOCAL: {  /* stat -> localstat */
      next(ls);  /* skip LOCAL */
      if (testnext(ls, TK_FUNCTION))  /* local function? */
        localfunc(ls);
      else
        localstat(ls);
      return 0;
    }
    case TK_RETURN: {  /* stat -> retstat */
      retstat(ls);
      return 1;  /* must be last statement */
    }
    case TK_BREAK: {  /* stat -> breakstat */
      breakstat(ls);
      return 1;  /* must be last statement */
    }
    case TK_DIVERTEDSTRING: { /* stat -> divertedstringstat */
      divertedstringstat(ls);
      return 0; /* Can be anywhere in a function */
    }
    default: {
      exprstat(ls);
      return 0;  /* to avoid warnings */
    }
  }
}


static void chunk (LexState *ls) {
  /* chunk -> { stat [`;'] } */
  int islast = 0;
  enterlevel(ls);
  while (!islast && !block_follow(ls->t.token)) {
    islast = statement(ls);
    testnext(ls, ';');
    lua_assert(ls->fs->f->maxstacksize >= ls->fs->freereg &&
               ls->fs->freereg >= ls->fs->nactvar);
    ls->fs->freereg = ls->fs->nactvar;  /* free registers */
  }
  leavelevel(ls);
}

/* }====================================================================== */
